
#include <module.h>
#include "amxxmodule.h"
#include <sm_trie_tpl.h>

static const int VirtualTableFunctionsNumber = 140;


#ifdef __linux__
	int VTable_offset = 0x94;
#else
	int VTable_offset = 0x0;
#endif

class VirtualTable
{
	void* vtable[VirtualTableFunctionsNumber];
	char classname_new[100];
	char classname_original[100];
	
	public:

		VirtualTable(char* classname_new,char* classname_original)
		{
			strncpy(this->classname_new,classname_new,(sizeof(this->classname_new)/sizeof(char)) - sizeof(char));
			strncpy(this->classname_original,classname_original,(sizeof(this->classname_original)/sizeof(char)) - sizeof(char));
		}

		void** get_vtable()
		{
			return vtable;
		}

		char* get_classname_original()
		{
			return this->classname_original;
		}

		char* get_classname_new()
		{
			return this->classname_new;
		}

		int get_classname_new_allocd()
		{
			return g_engfuncs.pfnAllocString(classname_new);
		}

		int get_classname_original_allocd()
		{
			return g_engfuncs.pfnAllocString(classname_original);
		}
};

KTrie<VirtualTable*> new_classnames;

VirtualTable* create_virtual_table(char* classname_new,char* classname_original)
{
	edict_t *entity = CREATE_ENTITY();
	CALL_GAME_ENTITY(PLID,classname_original,&entity->v);

	if(!entity->pvPrivateData)
	{
		REMOVE_ENTITY(entity);
		return NULL;
	}

	VirtualTable* virtualTable = new VirtualTable(classname_new,classname_original);

	void** vtable = *((void***)(((char*)entity->pvPrivateData)+VTable_offset));

	void** new_vtable = virtualTable->get_vtable();

	memcpy(new_vtable,vtable,sizeof(void*) * VirtualTableFunctionsNumber);

	REMOVE_ENTITY(entity);

	return virtualTable;
}

LibraryForModule* ModLibrary = NULL;

void OnLoadLibrary(LibraryForModule* library)
{
	if(!strcmp(library->get_name(),"mod"))
	{
		ModLibrary = library;
	}
}

VirtualTable* currentVirtualTable = NULL;

RageHookReturn OnCallGameEntityPre(plid_t*& plid, const char*& entStr,entvars_t*& pev,qboolean& ret)
{
	currentVirtualTable = NULL;

	VirtualTable** virtualTable_ptr = new_classnames.retrieve(entStr);

	if(virtualTable_ptr)
	{
		currentVirtualTable = *virtualTable_ptr;

		entStr = currentVirtualTable->get_classname_original();
	}

	return RageHookIgnored;
}

RageHookReturn OnCallGameEntityPost(plid_t*& plid, const char*& entStr,entvars_t*& pev,qboolean& ret)
{
	if(ret && currentVirtualTable)
	{
		void*** vtable_address = ((void***)(((char*)Rage->entvar_to_edict(pev)->pvPrivateData)+VTable_offset));
		*vtable_address = currentVirtualTable->get_vtable();

		pev->classname = currentVirtualTable->get_classname_new_allocd();
	}
	
	return RageHookIgnored;
}

RageHookReturn OnCreateNamedEntityPre(int& n,edict_t*& e)
{
	currentVirtualTable = NULL;

	VirtualTable** virtualTable_ptr = new_classnames.retrieve(g_engfuncs.pfnSzFromIndex(n));

	if(virtualTable_ptr)
	{
		currentVirtualTable = *virtualTable_ptr;

		n = currentVirtualTable->get_classname_original_allocd();
	}

	return RageHookIgnored;
}

RageHookReturn OnCreateNamedEntityPost(int& n,edict_t*& e)
{
	if(e && currentVirtualTable)
	{
		void*** vtable_address = ((void***)(((char*)e->pvPrivateData)+VTable_offset));
		*vtable_address = currentVirtualTable->get_vtable();

		e->v.classname = currentVirtualTable->get_classname_new_allocd();
	}

	return RageHookIgnored;
}

void OnLoadFunction( FunctionForModule* function )
{
	if( !strcmp( function->get_name(), "CreateNamedEntity" ) )
	{
		function->add_module_hook( ( void* )&OnCreateNamedEntityPre, RageHookPre );
		function->add_module_hook( ( void* )&OnCreateNamedEntityPost, RageHookPost );
	}
	else if( !strcmp( function->get_name(), "CallGameEntity" ) )
	{
		function->add_module_hook( ( void* )&OnCallGameEntityPre, RageHookPre );
		function->add_module_hook( ( void* )&OnCallGameEntityPost, RageHookPost );
	}
}


cell CreateNewClass(AMX* amx,cell* params)
{
	int len;
	char* classname_new_ = g_fn_GetAmxString(amx,params[1],0,&len);

	static char classname_new[100];
	strncpy(classname_new,classname_new_,len);
	classname_new[len] = 0;

	int len2;
	char* classname_original_ = g_fn_GetAmxString(amx,params[2],0,&len2);

	static char classname_original[100];
	strncpy(classname_original,classname_original_,len2);
	classname_original[len2] = 0;

	VirtualTable** virtual_table_ptr = new_classnames.retrieve(classname_new);

	if(virtual_table_ptr)
	{
		VirtualTable* virtual_table = *virtual_table_ptr;

		if(strcmp(classname_original,virtual_table->get_classname_original()))
		{
			MF_LogError(amx, AMX_ERR_NATIVE, "Tried to create '%s' based on %s but it was already created based on %s",virtual_table->get_classname_new(),classname_original,virtual_table->get_classname_original());
			return 0;
		}

		return 1;
	}
	
	VirtualTable* virtual_table = create_virtual_table(classname_new,classname_original);

	if(!virtual_table)
	{
		MF_LogError(amx, AMX_ERR_NATIVE, "Unable to create an entity of classname '%s'",classname_original);
		return 0;
	}

	new_classnames.insert(virtual_table->get_classname_new(),virtual_table);

	return 1;
}

AMX_NATIVE_INFO Exports[] =
{
	{"CreateNewClass",&CreateNewClass},
	{NULL, NULL}
};

void OnAmxxAttach()
{
	MF_AddNatives(Exports);
}

void OnLoadFunctionIncomplete(FunctionIncompleteForModule* function_incomplete)
{

}

void OnLoadVirtualFunction(VirtualFunctionForModule* virtual_function)
{

	
}
